@@ -1,4 +1,4 @@
-# Copyright 2021 The Flax Authors.
+# Copyright 2022 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -58,3 +58,6 @@ flax_filter_frames = bool_env('FLAX_FILTER_FRAMES', True)
 
 # Whether to automatically wrap Module methods with named_call for profiles.
 flax_profile = bool_env('FLAX_PROFILE', False)
+
+# Whether to use the lazy rng implementation
+flax_lazy_rng = bool_env('FLAX_LAZY_RNG', False)
