@@ -1,4 +1,4 @@
-# Copyright 2021 The Flax Authors.
+# Copyright 2022 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -12,19 +12,6 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-# Copyright 2020 The Flax Authors.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
-#
-#     http://www.apache.org/licenses/LICENSE-2.0
-#
-# Unless required by applicable law or agreed to in writing, software
-# distributed under the License is distributed on an "AS IS" BASIS,
-# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-# See the License for the specific language governing permissions and
-# limitations under the License.
 """"""  # Use an empty top-level docstring so Sphinx won't output the one below.
 """Flax error classes.
 
@@ -173,6 +160,23 @@ class ScopeParamNotFoundError(FlaxError):
                      f'"{scope_path}".')
 
 
+class ScopeCollectionNotFound(FlaxError):
+  """
+  This error is thrown when trying to access a variable from an empty collection.
+
+  There are two common causes:
+  1. The collection was not passed to ``apply`` correctly.
+    For example, you might have used ``module.apply(params, ...)`` instead
+    of ``module.apply({'params': params}, ...)``.
+  2. The collection is empty because the variables need to be initialized.
+    In this case, you should have made the collection mutable during
+    apply (e.g.: ``module.apply(variables, ..., mutable=['state'])``.
+  """
+  def __init__(self, col_name, var_name, scope_path):
+    super().__init__(f'Tried to access "{var_name}" from collection "{col_name}"" in '
+                     f'"{scope_path}" but the collection is emtpy.')
+
+
 class ScopeParamShapeError(FlaxError):
   """
   This error is thrown when the shape of an existing parameter is different from
@@ -523,7 +527,7 @@ class InvalidCheckpointError(FlaxError):
   A checkpoint cannot be stored in a directory that already has
   a checkpoint at the current or a later step.
 
-  You can pass `overwrite=True` to disable this behavior and
+  You can pass ``overwrite=True`` to disable this behavior and
   overwrite existing checkpoints in the target directory.
   """
   def __init__(self, path, step):
@@ -538,9 +542,10 @@ class TransformedMethodReturnValueError(FlaxError):
   """
   Transformed Module methods cannot return other Modules or Variables.
 
-  This commonly occurs when named_call is automatically applied to helper
-  constructor methods when profiling is enabled, and can be mitigated by
-  using the @nn.nowrap decorator to prevent automatic wrapping.
+  This commonly occurs when ``@nn.named_call`` is automatically applied to
+  helper constructor methods when profiling is enabled (``FLAX_PROFILE=true``
+  environment variable or via ``nn.enable_named_call()``), and can be mitigated
+  by using the ``@nn.nowrap`` decorator to prevent automatic wrapping.
   """
   def __init__(self, name):
     super().__init__(
