@@ -1,4 +1,4 @@
-# Copyright 2021 The Flax Authors.
+# Copyright 2022 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -108,19 +108,28 @@ def _parse_spec(spec):
 
 
 def prefetch_to_device(iterator, size, devices=None):
-  """"Shard and prefetch batches on device.
+  """Shard and prefetch batches on device.
 
   This utility takes an iterator and returns a new iterator which fills an on
   device prefetch buffer. Eager prefetching can improve the performance of
   training loops significantly by overlapping compute and data transfer.
 
-  This utility is mostly useful for GPUs, for TPUs it should not be necessary.
+  This utility is mostly useful for GPUs, for TPUs and CPUs it should not be
+  necessary -- the TPU & CPU memory allocators (normally) don't pick a memory
+  location that isn't free yet so they don't block. Instead those allocators OOM.
 
   Args:
     iterator: an iterator that yields a pytree of ndarrays where the first
       dimension is sharded across devices.
+
     size: the size of the prefetch buffer.
+
+      If you're training on GPUs, 2 is generally the best choice because this
+      guarantees that you can overlap a training step on GPU with a data
+      prefetch step on CPU.
+
     devices: the list of devices to which the arrays should be prefetched.
+
   Yields:
     The original items from the iterator where each ndarray is now a sharded to
     the specified devices.
@@ -150,7 +159,7 @@ def prefetch_to_device(iterator, size, devices=None):
     enqueue(1)
 
 
-def _scan_nd(body_fn, init, xs, n=1):
+def _scan_nd(body_fn, init, xs, n=1, unroll=(1,)):
   """Utility for performing an n-dimensional `lax.scan`.
 
   The n-d scan is simply recursive call of 1-d scan.
@@ -163,11 +172,11 @@ def _scan_nd(body_fn, init, xs, n=1):
     A tuple of the final carry and the values returned by the body.
   """
   if n == 1:
-    return lax.scan(body_fn, init, xs)
+    return lax.scan(body_fn, init, xs, unroll=unroll[0])
   else:
     def scan_body(c, x):
-      return _scan_nd(body_fn, c, x, n=n-1)
-    return lax.scan(scan_body, init, xs)
+      return _scan_nd(body_fn, c, x, n=n-1, unroll=unroll[1:])
+    return lax.scan(scan_body, init, xs, unroll=unroll[0])
 
 
 def _invert_perm(perm):
@@ -177,22 +186,38 @@ def _invert_perm(perm):
   return tuple(perm_inv)
 
 
-def scan_in_dim(body_fn, init, xs, axis=(0,), keepdims=False):
+def scan_in_dim(body_fn, init, xs, axis=(0,), unroll=(1,), keepdims=False):
   """utility for doing a scan along arbitrary dimensions.
 
-  see `lax.scan` for details on how the scan operation works.
+  See `lax.scan` for details on how the scan operation works.
+
+  Note on `unroll`: This argument gets left padded with ones to match the size
+  of `axis`. Doing so allows unrolls to performed from the innermost loop first.
+  For example, `scan_in_dim(..., axis=(1, 2, 3), unroll=5)` is equivalent to
+  `scan_in_dim(..., axis=(1, 2, 3), unroll=(1, 1, 5))`.
+
   Args:
     body_fn: the body of the loop of type (c, x) -> (c, y).
     init: initial value for the carry.
     xs: a pytree of tensors to scan over.
     axis: the axis to scan over.
     keepdims: keep the dimensions that are scanned over.
+    unroll: an optional positive integer, or tuple of positive integers
+      showing how many iterations of the loop to be unroll into a single
+      iteration for each axis.
   Returns:
     A tuple of the final carry and the values returned by the body.
   """
   if not isinstance(axis, Iterable):
     axis = (axis,)
 
+  if not isinstance(unroll, Iterable):
+    unroll = (unroll,)
+
+  # Pad unroll with ones so we start unrolling from the innermost loop
+  len_diff = len(axis) - len(unroll)
+  unroll = (1,) * len_diff + unroll
+
   def transpose_in(x):
     perm = axis + tuple(np.delete(np.arange(x.ndim), axis))
     return x.transpose(perm)
@@ -211,6 +236,6 @@ def scan_in_dim(body_fn, init, xs, axis=(0,), keepdims=False):
     return c, ys
 
   xs = jax.tree_map(transpose_in, xs)
-  c, ys = _scan_nd(body_wrapper, init, xs, n=len(axis))
+  c, ys = _scan_nd(body_wrapper, init, xs, n=len(axis), unroll=unroll)
   ys = jax.tree_map(transpose_out, ys)
   return c, ys
