@@ -1,4 +1,4 @@
-# Copyright 2021 The Flax Authors.
+# Copyright 2022 The Flax Authors.
 #
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
@@ -50,23 +50,25 @@ def _canonicalize_tuple(x):
 class DenseGeneral(Module):
   """A linear transformation with flexible axes.
 
-    Attributes:
-      features: int or tuple with number of output features.
-      axis: int or tuple with axes to apply the transformation on. For instance,
-        (-2, -1) will apply the transformation to the last two axes.
-      batch_dims: tuple with batch axes.
-      use_bias: whether to add a bias to the output (default: True).
-      dtype: the dtype of the computation (default: float32).
-      kernel_init: initializer function for the weight matrix.
-      bias_init: initializer function for the bias.
-      precision: numerical precision of the computation see `jax.lax.Precision`
-        for details.
+  Attributes:
+    features: int or tuple with number of output features.
+    axis: int or tuple with axes to apply the transformation on. For instance,
+      (-2, -1) will apply the transformation to the last two axes.
+    batch_dims: tuple with batch axes.
+    use_bias: whether to add a bias to the output (default: True).
+    dtype: the dtype of the computation (default: float32).
+    param_dtype: the dtype passed to parameter initializers (default: float32).
+    kernel_init: initializer function for the weight matrix.
+    bias_init: initializer function for the bias.
+    precision: numerical precision of the computation see `jax.lax.Precision`
+      for details.
   """
   features: Union[int, Iterable[int]]
   axis: Union[int, Iterable[int]] = -1
   batch_dims: Iterable[int] = ()
   use_bias: bool = True
   dtype: Dtype = jnp.float32
+  param_dtype: Dtype = jnp.float32
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
   precision: Any = None
@@ -112,7 +114,8 @@ class DenseGeneral(Module):
         inputs.shape[ax] if ax in batch_dims else 1
         for ax in range(inputs.ndim) if ax not in axis)
     kernel_shape = tuple([inputs.shape[ax] for ax in axis]) + features
-    kernel = self.param('kernel', kernel_init_wrap, batch_shape + kernel_shape)
+    kernel = self.param('kernel', kernel_init_wrap, batch_shape + kernel_shape,
+                        self.param_dtype)
     kernel = jnp.asarray(kernel, self.dtype)
 
     batch_ind = tuple(range(n_batch_dims))
@@ -130,7 +133,8 @@ class DenseGeneral(Module):
                                 for _ in range(size_batch_dims)], axis=0)
         return jnp.reshape(bias, shape)
 
-      bias = self.param('bias', bias_init_wrap, batch_shape + features)
+      bias = self.param('bias', bias_init_wrap, batch_shape + features,
+                        self.param_dtype)
       # expand bias shape to broadcast bias over batch dims.
       bias = jnp.reshape(bias, expanded_batch_shape + features)
       bias = jnp.asarray(bias, self.dtype)
@@ -145,6 +149,7 @@ class Dense(Module):
     features: the number of output features.
     use_bias: whether to add a bias to the output (default: True).
     dtype: the dtype of the computation (default: float32).
+    param_dtype: the dtype passed to parameter initializers (default: float32).
     precision: numerical precision of the computation see `jax.lax.Precision`
       for details.
     kernel_init: initializer function for the weight matrix.
@@ -152,7 +157,8 @@ class Dense(Module):
   """
   features: int
   use_bias: bool = True
-  dtype: Any = jnp.float32
+  dtype: Dtype = jnp.float32
+  param_dtype: Dtype = jnp.float32
   precision: Any = None
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
@@ -170,13 +176,15 @@ class Dense(Module):
     inputs = jnp.asarray(inputs, self.dtype)
     kernel = self.param('kernel',
                         self.kernel_init,
-                        (inputs.shape[-1], self.features))
+                        (inputs.shape[-1], self.features),
+                        self.param_dtype)
     kernel = jnp.asarray(kernel, self.dtype)
     y = lax.dot_general(inputs, kernel,
                         (((inputs.ndim - 1,), (0,)), ((), ())),
                         precision=self.precision)
     if self.use_bias:
-      bias = self.param('bias', self.bias_init, (self.features,))
+      bias = self.param('bias', self.bias_init, (self.features,),
+                        self.param_dtype)
       bias = jnp.asarray(bias, self.dtype)
       y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
     return y
@@ -201,9 +209,10 @@ class Conv(Module):
       be a sequence of integers.
     strides: an integer or a sequence of `n` integers, representing the
       inter-window strides (default: 1).
-    padding: either the string `'SAME'`, the string `'VALID'`, or a sequence
-      of `n` `(low, high)` integer pairs that give the padding to apply before
-      and after each spatial dimension.
+    padding: either the string `'SAME'`, the string `'VALID'`, the string
+      `'CIRCULAR'` (periodic boundary conditions), or a sequence of `n` `(low,
+      high)` integer pairs that give the padding to apply before and after each
+      spatial dimension.
     input_dilation: an integer or a sequence of `n` integers, giving the
       dilation factor to apply in each spatial dimension of `inputs` (default: 1).
       Convolution with input dilation `d` is equivalent to transposed
@@ -216,6 +225,7 @@ class Conv(Module):
       features into groups.
     use_bias: whether to add a bias to the output (default: True).
     dtype: the dtype of the computation (default: float32).
+    param_dtype: the dtype passed to parameter initializers (default: float32).
     precision: numerical precision of the computation see `jax.lax.Precision`
       for details.
     kernel_init: initializer for the convolutional kernel.
@@ -230,6 +240,7 @@ class Conv(Module):
   feature_group_count: int = 1
   use_bias: bool = True
   dtype: Dtype = jnp.float32
+  param_dtype: Dtype = jnp.float32
   precision: Any = None
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
@@ -237,9 +248,13 @@ class Conv(Module):
   @compact
   def __call__(self, inputs: Array) -> Array:
     """Applies a convolution to the inputs.
-
+ 
     Args:
       inputs: input data with dimensions (batch, spatial_dims..., features).
+        This is the channels-last convention, i.e. NHWC for a 2d convolution
+        and NDHWC for a 3D convolution. Note: this is different from the input
+        convention used by `lax.conv_general_dilated`, which puts the spatial
+        dimensions last.
 
     Returns:
       The convolved data.
@@ -275,15 +290,23 @@ class Conv(Module):
     assert in_features % self.feature_group_count == 0
     kernel_shape = kernel_size + (
         in_features // self.feature_group_count, self.features)
-    kernel = self.param('kernel', self.kernel_init, kernel_shape)
+    kernel = self.param('kernel', self.kernel_init, kernel_shape, self.param_dtype)
     kernel = jnp.asarray(kernel, self.dtype)
 
+    if self.padding == 'CIRCULAR':
+      kernel_size_dilated = [(k - 1) * d + 1 for k, d in zip(kernel_size, kernel_dilation)]
+      pads = [(0, 0)] + [((k - 1) // 2, k // 2) for k in kernel_size_dilated] + [(0, 0)]
+      inputs = jnp.pad(inputs, pads, mode='wrap')
+      padding_lax = 'VALID'
+    else:
+      padding_lax = self.padding
+
     dimension_numbers = _conv_dimension_numbers(inputs.shape)
     y = lax.conv_general_dilated(
         inputs,
         kernel,
         strides,
-        self.padding,
+        padding_lax,
         lhs_dilation=input_dilation,
         rhs_dilation=kernel_dilation,
         dimension_numbers=dimension_numbers,
@@ -293,7 +316,7 @@ class Conv(Module):
     if is_single_input:
       y = jnp.squeeze(y, axis=0)
     if self.use_bias:
-      bias = self.param('bias', self.bias_init, (self.features,))
+      bias = self.param('bias', self.bias_init, (self.features,), self.param_dtype)
       bias = jnp.asarray(bias, self.dtype)
       y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
     return y
@@ -307,17 +330,18 @@ class ConvTranspose(Module):
     kernel_size: shape of the convolutional kernel. For 1D convolution,
       the kernel size can be passed as an integer. For all other cases, it must
       be a sequence of integers.
-    strides: a sequence of `n` integers, representing the inter-window
-      strides.
-    padding: either the string `'SAME'`, the string `'VALID'`, or a sequence
-      of `n` `(low, high)` integer pairs that give the padding to apply before
-      and after each spatial dimension.
+    strides: a sequence of `n` integers, representing the inter-window strides.
+    padding: either the string `'SAME'`, the string `'VALID'`, the string
+      `'CIRCULAR'` (periodic boundary conditions), or a sequence of `n` `(low,
+      high)` integer pairs that give the padding to apply before and after each
+      spatial dimension.
     kernel_dilation: `None`, or a sequence of `n` integers, giving the
       dilation factor to apply in each spatial dimension of the convolution
       kernel. Convolution with kernel dilation is also known as 'atrous
       convolution'.
     use_bias: whether to add a bias to the output (default: True).
     dtype: the dtype of the computation (default: float32).
+    param_dtype: the dtype passed to parameter initializers (default: float32).
     precision: numerical precision of the computation see `jax.lax.Precision`
       for details.
     kernel_init: initializer for the convolutional kernel.
@@ -330,6 +354,7 @@ class ConvTranspose(Module):
   kernel_dilation: Optional[Iterable[int]] = None
   use_bias: bool = True
   dtype: Dtype = jnp.float32
+  param_dtype: Dtype = jnp.float32
   precision: Any = None
   kernel_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_kernel_init
   bias_init: Callable[[PRNGKey, Shape, Dtype], Array] = zeros
@@ -341,6 +366,10 @@ class ConvTranspose(Module):
 
     Args:
       inputs: input data with dimensions (batch, spatial_dims..., features).
+        This is the channels-last convention, i.e. NHWC for a 2d convolution
+        and NDHWC for a 3D convolution. Note: this is different from the input
+        convention used by `lax.conv_general_dilated`, which puts the spatial
+        dimensions last.
 
     Returns:
       The convolved data.
@@ -361,20 +390,56 @@ class ConvTranspose(Module):
 
     in_features = inputs.shape[-1]
     kernel_shape = kernel_size + (in_features, self.features)
-    kernel = self.param('kernel', self.kernel_init, kernel_shape)
+    kernel = self.param('kernel', self.kernel_init, kernel_shape, self.param_dtype)
     kernel = jnp.asarray(kernel, self.dtype)
 
+    if self.padding == 'CIRCULAR':
+      padding_lax = 'VALID'
+    else:
+      padding_lax = self.padding
+
     y = lax.conv_transpose(inputs,
                            kernel,
                            strides,
-                           self.padding,
+                           padding_lax,
                            rhs_dilation=self.kernel_dilation,
                            precision=self.precision)
 
+    if self.padding == "CIRCULAR":
+      # For circular padding, we need to identify the size of the final output
+      # ("period") along each spatial dimension, pad each dimension to an
+      # integer number of periods, and wrap the array periodically around each
+      # dimension. Padding should be done in such a way that the start of the
+      # original input data inside the padded array is located at integer
+      # number of periods - otherwise the result would be circularly shifted.
+
+      # Compute period along each spatial dimension - it's input size scaled
+      # by the stride.
+      scaled_x_dims = [
+        x_dim * stride for x_dim, stride in zip(inputs.shape[1:-1], strides)
+      ]
+      # Compute difference between the current size of y and the final output
+      # size, and complement this difference to 2 * period - that gives how
+      # much we need to pad.
+      size_diffs = [
+        -(y_dim - x_dim) % (2 * x_dim)
+        for y_dim, x_dim in zip(y.shape[1:-1], scaled_x_dims)
+      ]
+      # Divide the padding equaly between left and right. The choice to put
+      # "+1" on the left (and not on the right) represents a convention for
+      # aligning even-sized kernels.
+      total_pad = [((size_diff + 1) // 2, size_diff // 2) for size_diff in size_diffs]
+      y = np.pad(y, [(0, 0)] + total_pad + [(0, 0)])
+      # Wrap the result periodically around each spatial dimension,
+      # one by one.
+      for i in range(1, y.ndim - 1):
+        y = y.reshape(y.shape[:i] + (-1, scaled_x_dims[i - 1]) + y.shape[i + 1:])
+        y = y.sum(axis=i)
+
     if is_single_input:
       y = jnp.squeeze(y, axis=0)
     if self.use_bias:
-      bias = self.param('bias', self.bias_init, (self.features,))
+      bias = self.param('bias', self.bias_init, (self.features,), self.param_dtype)
       bias = jnp.asarray(bias, self.dtype)
       y += jnp.reshape(bias, (1,) * (y.ndim - 1) + (-1,))
     return y
@@ -392,11 +457,13 @@ class Embed(Module):
     num_embeddings: number of embeddings.
     features: number of feature dimensions for each embedding.
     dtype: the dtype of the embedding vectors (default: float32).
+    param_dtype: the dtype passed to parameter initializers (default: float32).
     embedding_init: embedding initializer.
   """
   num_embeddings: int
   features: int
   dtype: Dtype = jnp.float32
+  param_dtype: Dtype = jnp.float32
   embedding_init: Callable[[PRNGKey, Shape, Dtype], Array] = default_embed_init
 
   embedding: Array = field(init=False)
@@ -405,7 +472,7 @@ class Embed(Module):
     self.embedding = self.param('embedding',
                                 self.embedding_init,
                                 (self.num_embeddings, self.features),
-                                self.dtype)
+                                self.param_dtype)
 
   def __call__(self, inputs):
     """Embeds the inputs along the last dimension.
@@ -420,7 +487,8 @@ class Embed(Module):
     if not jnp.issubdtype(inputs.dtype, jnp.integer):
       raise ValueError('Input type must be an integer or unsigned integer.')
     # Use take because fancy indexing numpy arrays with JAX indices does not work correctly.
-    return jnp.take(self.embedding, inputs, axis=0)
+    embedding = jnp.asarray(self.embedding, self.dtype)
+    return jnp.take(embedding, inputs, axis=0)
 
   def attend(self, query):
     """Attend over the embedding using a query array.
@@ -434,4 +502,6 @@ class Embed(Module):
       Commonly used for weight-sharing between embeddings and logit transform
       in NLP models.
     """
-    return jnp.dot(query, self.embedding.T)
+    query = jnp.asarray(query, self.dtype)
+    embedding = jnp.asarray(self.embedding, self.dtype)
+    return jnp.dot(query, embedding.T)
